# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""pytest plugin for dasf setups."""

from __future__ import annotations

import json
import os
import subprocess as spr
import sys
import time
import uuid
from typing import TYPE_CHECKING, Callable, Dict, Generator, List, Union

import pytest
from pytest_django.lazy_django import skip_if_no_django

if TYPE_CHECKING:
    from .live_ws_server_helper import ChannelsLiveServer


@pytest.fixture
def live_ws_server(
    transactional_db, request: pytest.FixtureRequest
) -> Generator[ChannelsLiveServer, None, None]:
    """Run a live Daphne server in the background during tests

    The address the server is started from is taken from the
    --liveserver command line option or if this is not provided from
    the DJANGO_LIVE_TEST_SERVER_ADDRESS environment variable.  If
    neither is provided ``localhost`` is used.  See the Django
    documentation for its full syntax.

    Notes
    -----
    - Static assets will be automatically served when
      ``django.contrib.staticfiles`` is available in INSTALLED_APPS.
    - this fixture is a combination of the ``live_server`` fixture of
      pytest-django and channels ``ChannelsLiveServerTestCase``. Different
      from pytest-djangos fixture, we need a function-scope however, as the
      daphne process runs in a different process (and not in a different
      thread as with pytest-django).
    - You can't use an in-memory database for your live tests. Therefore
      include a test database file name in your settings to tell Django to use
      a file database if you use SQLite:

      .. code-block:: python

          DATABASES = {
              "default": {
                  "ENGINE": "django.db.backends.sqlite3",
                  "NAME": os.path.join(BASE_DIR, "db.sqlite3"),
                  "TEST": {
                      "NAME": os.path.join(BASE_DIR, "db_test.sqlite3"),
                  },
              },
          }
    """
    import dasf_broker.tests.live_ws_server_helper as live_ws_server_helper

    skip_if_no_django()

    server = live_ws_server_helper.ChannelsLiveServer("localhost")
    yield server
    server.stop()


def create_brokertopic(slug: str):
    from guardian.shortcuts import assign_perm, get_anonymous_user

    from dasf_broker.models import BrokerTopic

    topic = BrokerTopic.objects.create(slug=slug, is_public=True)
    assign_perm("can_consume", get_anonymous_user(), topic)


@pytest.fixture
def random_topic(db) -> str:
    """Generate a random topic for testing."""
    topic_slug = "test_topic_" + uuid.uuid4().urn[9:]
    create_brokertopic(topic_slug)
    return topic_slug


def get_test_module_args(
    topic: str,
    live_ws_server: ChannelsLiveServer,
) -> List[str]:
    """Generate a command for subprocess to launch the test module."""
    from dasf_broker import app_settings

    return [
        "-t",
        topic,
        "--websocket-url",
        "%s/%s"
        % (live_ws_server.ws_url, app_settings.DASF_WEBSOCKET_URL_ROUTE),
    ]


@pytest.fixture
def get_module_command(live_ws_server) -> Callable[[str, str], List[str]]:
    """A factory to generate commands from a backend module."""

    def factory(topic: str, path_or_mod: str) -> List[str]:
        """Generate a command to connect a script backend module."""
        if os.path.exists(path_or_mod):
            base = [sys.executable, path_or_mod]
        else:
            base = [sys.executable, "-m", path_or_mod]
        return base + get_test_module_args(topic, live_ws_server)

    return factory


@pytest.fixture
def connect_module(
    get_module_command: Callable[[str, str], List[str]],
) -> Generator[Callable[[str, str], spr.Popen], None, None]:
    """Get a factory that connects DASF backend module scripts"""
    processes: List[spr.Popen] = []

    def connect(topic: str, path_or_mod: str, *args, **kwargs) -> spr.Popen:
        """Connect a script to the live_server"""
        command = get_module_command(topic, path_or_mod)
        try:
            process = spr.Popen(command + ["listen"] + list(args), **kwargs)
        except Exception:
            raise
        else:
            processes.append(process)
            time.sleep(1)
            return process

    yield connect

    for process in processes:
        process.terminate()


@pytest.fixture
def test_dasf_connect(
    get_module_command: Callable[[str, str], List[str]],
) -> Callable[[str, str], str]:
    """Factory to test connecting a DASF module"""

    def test_connect(topic: str, path_or_mod: str, **kwargs) -> str:
        command = get_module_command(topic, path_or_mod)
        output = spr.check_output(command + ["test-connect"], **kwargs)
        return output.decode("utf-8")

    return test_connect


@pytest.fixture
def test_dasf_request(
    get_module_command: Callable[[str, str], List[str]], tmpdir
) -> Callable[[str, str, Union[Dict, str]], str]:
    """A factory to send requests via DASF.

    Notes
    -----
    Sending a request requires a connected backend module! See the fixture
    :func:`connect_module`.
    """

    def test_request(
        topic: str, path_or_mod: str, request: Union[Dict, str], **kwargs
    ) -> str:
        """Send a DASF request to a backend module."""
        command = get_module_command(topic, path_or_mod)
        request_path: str
        if not isinstance(request, str):
            request_path = str(tmpdir.mkdir(topic).join("request.json"))
            with open(request_path, "w") as f:
                json.dump(request, f)
        else:
            request_path = request

        output = spr.check_output(
            command + ["send-request", request_path], **kwargs
        )
        return output.decode("utf-8")

    return test_request



"""A module for a live_server with support for django-channels."""

# SPDX-FileCopyrightText: 2024 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: Apache-2.0

from functools import partial
from typing import Any, Dict

from channels.routing import get_default_application
from daphne.testing import DaphneProcess
from django.contrib.staticfiles.handlers import ASGIStaticFilesHandler
from django.core.exceptions import ImproperlyConfigured
from django.db import connections
from django.test.utils import modify_settings


def make_application(*, static_wrapper):
    # Module-level function for pickle-ability
    application = get_default_application()
    if static_wrapper is not None:
        application = static_wrapper(application)
    return application


class ChannelsLiveServer:
    """
    Does basically the same as TransactionTestCase but also launches a
    live Daphne server in a separate process, so
    that the tests may use another test framework, such as Selenium,
    instead of the built-in dummy client.
    """

    host = "localhost"
    ProtocolServerProcess = DaphneProcess
    static_wrapper = ASGIStaticFilesHandler
    serve_static = True

    def __init__(self, addr: str, *, start: bool = True) -> None:
        self.live_server_kwargs: Dict[str, Any] = {}
        try:
            host, port = addr.split(":")
        except ValueError:
            host = addr
        else:
            self.live_server_kwargs["port"] = int(port)

        for connection in connections.all():
            if self._is_in_memory_db(connection):
                raise ImproperlyConfigured(
                    "ChannelLiveServerTestCase can not be used with in memory databases"
                )

        self.host = host

        self._live_server_modified_settings = modify_settings(
            ALLOWED_HOSTS={"append": self.host}
        )

        self._live_server_modified_settings.enable()

        get_application = partial(
            make_application,
            static_wrapper=self.static_wrapper if self.serve_static else None,
        )
        self._server_process = self.ProtocolServerProcess(
            self.host, get_application, **self.live_server_kwargs
        )

        if start:
            self._server_process.start()
            self._server_process.ready.wait()
        self._port = self._server_process.port.value

    @property
    def url(self):
        return "http://%s:%s" % (self.host, self._port)

    @property
    def ws_url(self):
        return "ws://%s:%s" % (self.host, self._port)

    def stop(self):
        self._live_server_modified_settings.disable()
        self._server_process.terminate()
        self._server_process.join()

    def _is_in_memory_db(self, connection):
        """
        Check if DatabaseWrapper holds in memory database.
        """
        if connection.vendor == "sqlite":
            return connection.is_in_memory_db()



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Views
-----

Views of the dasf-broker-django app to be imported via the url
config (see :mod:`dasf_broker.urls`).
"""

from __future__ import annotations

from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import redirect
from django.views import generic  # noqa: F401
from guardian.mixins import PermissionRequiredMixin

from dasf_broker import models  # noqa: F401


class HttpResponseServiceUnavailable(HttpResponse):
    status_code = 503


class BrokerTopicStatusView(
    PermissionRequiredMixin, generic.detail.BaseDetailView
):
    """Get a hint on the status of a broker topic."""

    model = models.BrokerTopic

    permission_required = [
        "dasf_broker.can_view_status",
        "dasf_broker.can_produce",
        "dasf_broker.can_consume",
    ]

    any_perm = True

    accept_global_perms = True

    def render_to_response(self, context):
        availability = self.object.availability

        if not self.object.supports_dasf:
            return HttpResponseBadRequest("Topic is exluded from ping")
        elif availability is None:
            return HttpResponseServiceUnavailable("Status unknown")
        elif not availability:
            return HttpResponseServiceUnavailable(
                "Consumer could not be reached"
            )
        else:
            return HttpResponse("Consumer connected")

    def check_permissions(self, request):
        topic = self.get_permission_object()
        if topic.is_public:
            return
        return super().check_permissions(request)


class BrokerTopicPingView(
    PermissionRequiredMixin, generic.detail.SingleObjectMixin, generic.View
):
    """View to ping a broker topic."""

    model = models.BrokerTopic

    permission_required = "dasf_broker.can_produce"

    accept_global_perms = True

    def check_permissions(self, request):
        topic = self.get_permission_object()
        if topic.is_public:
            return
        return super().check_permissions(request)

    def get(self, request, *args, **kwargs):
        return self.post(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        topic = self.get_object()
        topic.ping()
        return redirect("dasf_broker:brokertopic-status", self.kwargs["slug"])



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Models
------

Models for the dasf-broker-django app.
"""

from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING, Dict, Optional

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.db import models  # noqa: F401
from django.utils.functional import cached_property
from django.utils.timezone import now
from guardian.shortcuts import assign_perm, get_users_with_perms

from dasf_broker import app_settings  # noqa: F401
from dasf_broker.constants import PropertyKeys

if TYPE_CHECKING:
    from django.contrib.auth.models import User


class BrokerTopicQuerySet(models.QuerySet):
    """A queryset for broker topics."""

    def filter_online(self, *args, **kwargs):
        """Query all online broker topics."""
        if args:
            args = list(args)
            args[0] &= (
                models.Q(supports_dasf=True)
                & models.Q(last_ping__gte=now() - dt.timedelta(minutes=2))
                & models.Q(last_pong__gte=models.F("last_ping"))
            )
        else:
            kwargs.update(
                dict(
                    last_pong__gte=models.F("last_ping"),
                    last_ping__gte=now() - dt.timedelta(minutes=2),
                )
            )
        return self.filter(*args, **kwargs)

    def filter_offline(self, *args):
        """Query all online broker topics."""
        args = list(args)
        query = (
            models.Q(supports_dasf=True)
            & models.Q(last_ping__gte=now() - dt.timedelta(minutes=2))
            & (
                models.Q(last_pong__isnull=True)
                | models.Q(
                    last_pong__lte=models.F("last_ping")
                    - dt.timedelta(seconds=10)
                )
            )
        )
        if args:
            args[0] &= query
        else:
            args = [query]
        return self.filter(*args)

    def filter_unknown_availability(self, *args):
        """Query all topics where the availability is unknown."""
        query = (
            models.Q(supports_dasf=False)
            | models.Q(last_ping__isnull=True)
            | models.Q(last_ping__lte=now() - dt.timedelta(minutes=2))
        )
        if args:
            args = list(args)
            args[0] &= query
        else:
            args = [query]
        return self.filter(*args)


class BrokerTopicManager(models.Manager.from_queryset(BrokerTopicQuerySet)):  # type: ignore
    """A manager for broker topics."""

    pass


class BrokerTopic(models.Model):
    """A topic for producing and consuming requests via websocket"""

    class Meta:
        permissions = (
            ("can_produce", "Can publish messages to the topic (Producer)."),
            ("can_consume", "Can consume messages to the topic (Consumer)."),
            ("can_view_status", "Can view the status of the consumer."),
        )

    class StoreMessageChoices(models.TextChoices):
        """Choices for storing messages."""

        DISABLED = "disabled", "Do not store messages at all"
        CACHE = (
            "cache",
            "Cache until it has been delivered to at least on consumer",
        )
        CACHEALL = (
            "cacheall",
            "Cache until it has been delivered to all consumers",
        )
        STORE = "store", "Store messages forever"

    objects = BrokerTopicManager()

    slug = models.SlugField(
        unique=True,
        help_text="Unique identifier for the topic.",
        db_index=True,
    )

    is_public = models.BooleanField(
        default=False, help_text="Can everyone publish data to this topic?"
    )

    date_created = models.DateTimeField(
        auto_now_add=True,
        help_text="The date and time when the message has been created",
    )

    last_ping = models.DateTimeField(
        null=True, blank=True, help_text="When has the topic last been pinged?"
    )

    last_pong = models.DateTimeField(
        null=True,
        blank=True,
        help_text="When has the topic last replyed on a ping?",
    )

    supports_dasf = models.BooleanField(
        default=True,
        help_text="Does this topic support the DASF messaging protocoll?",
    )

    store_messages = models.CharField(
        max_length=20,
        null=True,
        blank=True,
        help_text="Should messages be stored?",
        choices=[(None, "Use the server default")]
        + StoreMessageChoices.choices,  # type: ignore
    )

    garbage_collect_on = models.DateTimeField(
        help_text=(
            "Specify a time when this topic should be removed from the "
            "database."
        ),
        null=True,
        blank=True,
    )

    #: metadata field to allow additional attributes for topics.
    api_info = models.JSONField(default=dict, encoder=DjangoJSONEncoder)

    @property
    def consumers(self) -> models.QuerySet[User]:
        return get_users_with_perms(self, only_with_perms_in=["can_consume"])  # type: ignore[return-value]

    @property
    def producers(self) -> models.QuerySet[User]:
        return get_users_with_perms(self, only_with_perms_in=["can_produce"])  # type: ignore[return-value]

    @property
    def status_viewers(self) -> models.QuerySet[User]:
        return get_users_with_perms(  # type: ignore[return-value]
            self, only_with_perms_in=["can_view_status"]
        )

    @property
    def effective_store_messages(self) -> StoreMessageChoices:
        """Get the store message rule for this topic."""
        if self.store_messages is None:
            if self.is_response_topic:
                return self.StoreMessageChoices(
                    app_settings.DASF_STORE_RESPONSE_MESSAGES
                )
            else:
                return self.StoreMessageChoices(
                    app_settings.DASF_STORE_SOURCE_MESSAGES
                )
        else:
            return self.store_messages  # type: ignore

    @classmethod
    def build_websocket_url(cls, request, route: Optional[str] = None) -> str:
        base_route = app_settings.DASF_WEBSOCKET_URL_ROUTE
        if not base_route.endswith("/"):
            base_route = base_route + "/"

        path = "/" + base_route
        if route:
            path += route
        if getattr(settings, "FORCE_SCRIPT_NAME", None):
            path = settings.FORCE_SCRIPT_NAME + path  # type: ignore

        if app_settings.ROOT_URL:
            http_url = app_settings.ROOT_URL + path
        else:
            http_url = request.build_absolute_uri(path)
        ws_url = "ws" + http_url[4:]  # replace http with ws
        return ws_url

    def get_websocket_url(self, request) -> str:
        """Get the websocket url for this topic."""
        return self.build_websocket_url(request, self.slug)

    def create_and_send_message(self, user: User, content: Dict):
        """Create and send a message for the user"""
        message_id = content["messageId"]
        props = content.get("properties", {})

        consumers = self.consumers

        if PropertyKeys.RESPONSE_TOPIC in props:
            response_topic, created = ResponseTopic.objects.get_or_create(
                slug=props[PropertyKeys.RESPONSE_TOPIC],
                is_response_for=self,
                defaults=dict(supports_dasf=True),
            )
            if created:
                assign_perm(
                    "can_consume", user, response_topic.brokertopic_ptr
                )
                for consumer in consumers:
                    assign_perm(
                        "can_produce", consumer, response_topic.brokertopic_ptr
                    )
        else:
            response_topic = None

        date_created = now()
        content["publishTime"] = date_created.isoformat()

        if self.effective_store_messages != self.StoreMessageChoices.DISABLED:
            message = BrokerMessage.objects.create(
                message_id=message_id,
                topic=self,
                content=content,
                user=user,
                date_created=date_created,
                context=props.get("requestContext"),
            )
            assign_perm("view_brokermessage", user, message)
            for consumer in consumers:
                assign_perm("view_brokermessage", consumer, message)
            if response_topic is not None:
                response_topic.source_messages.add(message)

        channel_layer = get_channel_layer()

        async_to_sync(channel_layer.group_send)(
            f"dasf_topic_{self.slug}",
            {"type": "dasf.message", "content": content},
        )

    def ping(self):
        """Create a ping message and send it to the consumer."""
        if not self.supports_dasf:
            # do nothing
            return
        channel_layer = get_channel_layer()
        date_pinged = now()
        self.last_ping = date_pinged
        async_to_sync(channel_layer.group_send)(
            f"dasf_topic_{self.slug}",
            {"type": "dasf.ping"},
        )
        self.save()

    @cached_property
    def availability(self) -> Optional[bool]:
        """Get the online/offline status for the topic.

        This value can be ``True``, ``False`` or ``None``:

        None
            The status is unknown. This occurs when the last ping was more than
            two minutes ago or the topic has never, been pinged.
        False
            The was no pong yet or the last pong was before the last ping and
            the last ping was less than two minutes ago.
        True
            The topic is online, i.e. we received a pong after the last ping
            and the last ping was less then two minutes ago.
        """
        if (
            not self.supports_dasf
            or not self.last_ping
            or self.last_ping < now() - dt.timedelta(minutes=2)
        ):
            return None  # status unknown
        elif (
            not self.last_pong
            or self.last_pong < self.last_ping - dt.timedelta(seconds=10)
        ):
            return False  # status offline
        else:
            return True  # status online

    @property
    def is_response_topic(self) -> bool:
        """Is this topic a responsetopic?"""
        return hasattr(self, "responsetopic")

    def get_outstanding_messages(
        self, user: Optional[User] = None
    ) -> models.QuerySet[BrokerMessage]:
        """Get the messages that still need to be send.

        Parameters
        ----------
        user: Optional[User]
            The user for whom to send the messages. If None, the messages will
            be returned that have not yet been acknowledged at all.

        Returns
        -------
        models.QuerySet[BrokerMessage]
            A QuerySet of messages
        """
        if user is None:
            return self.brokermessage_set.filter(delivered_to=None)
        else:
            return self.brokermessage_set.filter(
                ~models.Q(delivered_to__pk=user.pk)
            )

    def __str__(self) -> str:
        return self.slug


class BrokerMessage(models.Model):
    """A message sent to the broker."""

    class Meta:
        constraints = [
            models.UniqueConstraint(
                name="unique_message_id_for_topic",
                fields=("message_id", "topic"),
            )
        ]

    message_id = models.UUIDField(help_text="Message ID", db_index=True)

    context = models.IntegerField(
        null=True,
        blank=True,
        help_text="Message context for messages from a producer.",
    )

    topic = models.ForeignKey(
        BrokerTopic,
        on_delete=models.CASCADE,
        help_text="The topic the message was published for.",
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        help_text="The user who produced the message.",
    )

    content = models.JSONField(help_text="The content of the message.")

    delivered_to = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        help_text="To whom has this message been delivered already?",
        related_name="brokermessage_set_delivered",
    )

    date_created = models.DateTimeField(
        help_text="The date and time when the message has been created",
    )

    def send(self):
        """Send the message via the websocket."""
        channel_layer = get_channel_layer()

        async_to_sync(channel_layer.group_send)(
            f"dasf_topic_{self.topic.slug}",
            {"type": "dasf.message", "content": self.content},
        )

    @property
    def delivered_to_all(self) -> bool:
        """Test if the message has been delivered to all consumers."""
        delivered_to = self.delivered_to.values_list("pk", flat=True)
        return not self.topic.consumers.filter(
            ~models.Q(pk__in=delivered_to)
        ).exists()

    @property
    def is_response(self) -> bool:
        """Is this message a response to a DASF request?"""
        return self.topic.is_response_topic

    def __str__(self) -> str:
        return f"Message {self.message_id} for topic {self.topic}"


class ResponseTopic(BrokerTopic):
    """A topic that accepts responses for messages."""

    brokertopic_ptr: BrokerTopic

    is_response_for = models.ForeignKey(
        BrokerTopic,
        null=True,
        blank=True,
        on_delete=models.CASCADE,
        related_name="responsetopics",
        help_text="Is this topic used only once for a single response?",
    )

    source_messages = models.ManyToManyField(
        BrokerMessage,
        blank=True,
        help_text="Messages from the producer to the reference topic.",
    )

    @property
    def is_response_topic(self) -> bool:
        """Is this topic a responsetopic?"""
        return True

    def __str__(self) -> str:
        return f"Response topic {super().__str__()}"



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""DASF Template tags
------------------

This module contains the template tags that should be used within DASF.
"""

from __future__ import annotations

import base64
import json
from typing import TYPE_CHECKING, Optional

from django import template

from dasf_broker import models

if TYPE_CHECKING:
    from dasf_broker.models import BrokerMessage

register = template.Library()


@register.simple_tag(takes_context=True)
def dasf_ws_url(context, route: Optional[str] = None) -> str:
    """Get the url route to a websocket.

    Parameters
    ----------
    route: Optional[str]
        The location of the route (i.e. the broker topic slug).

    Examples
    --------
    Call this tag in your template like::
        {% load dasf %}
        {% dasf_ws_url "test-topic" %}

    which will resolve to something like `ws://localhost:8000/ws/test-topic`
    (depending on your
    :attr:`~dasf_broker.app_settings.DASF_WEBSOCKET_URL_ROUTE`)
    """
    return models.BrokerTopic.build_websocket_url(context["request"], route)


@register.filter
def payload(message: BrokerMessage):
    payload = base64.b64decode(message.content.get("payload", "")).decode(
        "utf-8"
    )
    try:
        return json.loads(payload)
    except json.JSONDecodeError:
        return None


@register.inclusion_tag("dasf_broker/payload_str.html")
def payload_pre(message: BrokerMessage):
    """The payload as json formatted string"""
    payload = base64.b64decode(message.content.get("payload", "")).decode(
        "utf-8"
    )
    try:
        loaded = json.loads(payload)
    except json.JSONDecodeError:
        return {"payload": payload}
    else:
        return {"payload": json.dumps(loaded, indent=4)}



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

# Generated by Django 3.2.15 on 2023-01-09 20:42

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="BrokerTopic",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "slug",
                    models.SlugField(
                        help_text="Unique identifier for the topic.",
                        unique=True,
                    ),
                ),
                (
                    "is_public",
                    models.BooleanField(
                        default=False,
                        help_text="Can everyone publish data to this topic?",
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        auto_now_add=True,
                        help_text="The date and time when the message has been created",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "can_produce",
                        "Can publish messages to the topic (Producer).",
                    ),
                    (
                        "can_consume",
                        "Can consume messages to the topic (Consumer).",
                    ),
                ),
            },
        ),
        migrations.CreateModel(
            name="BrokerMessage",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "message_id",
                    models.UUIDField(db_index=True, help_text="Message ID"),
                ),
                (
                    "context",
                    models.IntegerField(
                        blank=True,
                        help_text="Message context for messages from a producer.",
                        null=True,
                    ),
                ),
                (
                    "content",
                    models.JSONField(help_text="The content of the message."),
                ),
                (
                    "delivered",
                    models.BooleanField(
                        default=False,
                        help_text="Has this message already been delivered?",
                    ),
                ),
                (
                    "date_created",
                    models.DateTimeField(
                        help_text="The date and time when the message has been created"
                    ),
                ),
                (
                    "topic",
                    models.ForeignKey(
                        help_text="The topic the message was published for.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="dasf_broker.brokertopic",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="The user who produced the message.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ResponseTopic",
            fields=[
                (
                    "brokertopic_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="dasf_broker.brokertopic",
                    ),
                ),
                (
                    "is_response_for",
                    models.ForeignKey(
                        blank=True,
                        help_text="Is this topic used only once for a single response?",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="responsetopics",
                        to="dasf_broker.brokertopic",
                    ),
                ),
                (
                    "source_messages",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Messages from the producer to the reference topic.",
                        to="dasf_broker.BrokerMessage",
                    ),
                ),
            ],
            bases=("dasf_broker.brokertopic",),
        ),
        migrations.AddConstraint(
            model_name="brokermessage",
            constraint=models.UniqueConstraint(
                fields=("message_id", "topic"),
                name="unique_message_id_for_topic",
            ),
        ),
    ]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Garbage collect DASF topics."""

from django.utils.timezone import now
from django_extensions.management.jobs import MinutelyJob


class Job(MinutelyJob):
    help = "Garbage collect DASF topics."

    def execute(self):
        from dasf_broker.models import BrokerTopic

        BrokerTopic.objects.filter(garbage_collect_on__lte=now()).delete()



# SPDX-FileCopyrightText: 2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

from django.apps import AppConfig


class DasfFrontendConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "dasf_broker.frontend"
    label = "dasf_frontend"



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Tests to connect a backend module."""

import subprocess as spr
from typing import Callable, Dict, Union


def test_hello_world_connect(
    test_dasf_connect: Callable[[str, str], str],
    random_topic: str,
    get_test_module_path: Callable[[str], str],
):
    """Test connecting the hello world backend module."""
    modpath = get_test_module_path("hello_world")
    test_dasf_connect(random_topic, modpath)


def test_dasf_request(
    connect_module: Callable[[str, str], spr.Popen],
    test_dasf_request: Callable[[str, str, Union[Dict, str]], str],
    random_topic: str,
    get_test_module_path: Callable[[str], str],
):
    """Test sending a request to the backend module."""

    modpath = get_test_module_path("hello_world")
    connect_module(random_topic, modpath)
    output = test_dasf_request(
        random_topic, modpath, {"func_name": "hello_world"}
    )

    assert "Hello World!" in output



# SPDX-FileCopyrightText: 2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

from typing import Any

from django.contrib.auth.mixins import UserPassesTestMixin
from django.views.generic import DetailView

from dasf_broker import app_settings, models


class BrokerTopicFrontendView(UserPassesTestMixin, DetailView):
    """A detail view for a broker topic."""

    model = models.BrokerTopic

    template_name_suffix = "_frontend"

    def test_func(self) -> bool:
        topic: models.BrokerTopic = self.get_object()
        user = self.request.user
        return (
            topic.is_public
            or user.has_perm("dasf_broker.can_produce")
            or user.has_perm("dasf_broker.can_produce", topic)
        )

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(responsetopic__isnull=True, supports_dasf=True)
        )

    def get_context_data(self, **kwargs) -> dict[str, Any]:
        context = super().get_context_data(**kwargs)
        context["ws_url"] = app_settings.DASF_WEBSOCKET_URL_ROUTE
        # the following may be overwritten by subclasses to add more params
        # to the `dasf-module` tag
        context["module_params"] = {}
        return context



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Test file for imports."""


def test_package_import():
    """Test the import of the main package."""
    import dasf_broker  # noqa: F401



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

from channels.auth import AuthMiddleware
from channels.db import database_sync_to_async
from channels.sessions import CookieMiddleware, SessionMiddleware


@database_sync_to_async
def get_user(scope):
    """
    Return the user model instance associated with the given scope.
    If no user is retrieved, return an instance of `AnonymousUser`.
    """
    from guardian.shortcuts import get_anonymous_user
    from rest_framework.authtoken.models import Token

    headers = dict(scope["headers"])
    if b"authorization" not in headers:
        raise ValueError(
            "Cannot find authorization in scope. You should wrap your "
            "consumer in TokenAuthMiddleware."
        )
    token_key = headers[b"authorization"].decode().split()[-1]
    try:
        token = Token.objects.get(key=token_key)
    except Token.DoesNotExist:
        return get_anonymous_user()
    else:
        return token.user


class TokenAuthMiddleware(AuthMiddleware):
    """
    Token authorization middleware for Django Channels
    """

    async def resolve_scope(self, scope):
        headers = dict(scope["headers"])
        if b"authorization" in headers:
            try:
                token_name, token_key = (
                    headers[b"authorization"].decode().split()
                )
            except ValueError:
                # other authorization method
                pass
            else:
                if token_name in ["Token", "Bearer"]:
                    scope["user"]._wrapped = await get_user(scope)
                    return
        return await super().resolve_scope(scope)


def TokenAuthMiddlewareStack(inner):
    return CookieMiddleware(SessionMiddleware(TokenAuthMiddleware(inner)))



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

from django.utils.log import AdminEmailHandler as BaseAdminEmailHandler


class AdminEmailHandler(BaseAdminEmailHandler):
    def emit(self, record):
        if record.status_code < 500:
            return
        else:
            if hasattr(record, "request") and hasattr(
                record.request, "resolver_match"
            ):
                match = record.request.resolver_match
                if (
                    match.url_name == "brokertopic-status"
                    and match.app_names == ["dasf_broker"]
                ):
                    return
            super().emit(record)



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Admin interfaces
----------------

This module defines the dasf-broker-django
Admin interfaces.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Tuple

from django.contrib import admin  # noqa: F401
from django.utils.timezone import now
from guardian.admin import GuardedModelAdmin

from dasf_broker import models  # noqa: F401

if TYPE_CHECKING:
    from django.db.models import QuerySet


class ResponseTopicExistsFilter(admin.SimpleListFilter):
    """Filter topics for response topics"""

    title = "is response topic"

    parameter_name = "responsetopic"

    def lookups(self, request: Any, model_admin: Any) -> List[Tuple[Any, str]]:
        return [("1", "Yes"), ("0", "No")]

    def queryset(
        self, request: Any, queryset: QuerySet[Any]
    ) -> Optional[QuerySet[Any]]:
        value = self.value()
        if value == "1":
            return queryset.filter(responsetopic__isnull=False)
        elif value == "0":
            return queryset.filter(responsetopic__isnull=True)
        else:
            return queryset


class AvailabilityFilter(admin.SimpleListFilter):
    """Filter topics based on their status."""

    title = "Availability"

    parameter_name = "last_pong"

    def lookups(self, request: Any, model_admin: Any) -> List[Tuple[Any, str]]:
        return [
            ("offline", "Offline"),
            ("online", "Online"),
            ("unknown", "Unknown"),
        ]

    def queryset(  # type: ignore
        self,
        request: Any,
        queryset: models.BrokerTopicQuerySet,  # type: ignore
    ) -> Optional[QuerySet[Any]]:
        value = self.value()
        if value == "offline":
            return queryset.filter_offline()
        elif value == "online":
            return queryset.filter_online()
        elif value == "unknown":
            return queryset.filter_unknown_availability()
        else:
            return queryset


@admin.action(description="Ping topics")
def ping_topics(modeladmin, request, queryset):
    """Ping the broker topics."""
    for topic in queryset.filter(supports_dasf=True):
        topic.ping()


@admin.action(description="Garbage collect topics")
def garbage_collect(modeladmin, request, queryset):
    queryset.filter(garbage_collect_on__lte=now()).delete()


@admin.register(models.BrokerTopic)
class BrokerTopicAdmin(GuardedModelAdmin):
    """An admin for :model:`dasf_broker.BrokerTopic"""

    search_fields = [
        "slug__icontains",
    ]

    list_display = [
        "slug",
        "get_consumers",
        "get_producers",
        "is_public",
        "is_response_topic",
        "get_availability",
        "pending_messages",
        "stored_messages",
        "stored_responses",
    ]

    list_filter = [
        "is_public",
        "date_created",
        ResponseTopicExistsFilter,
        AvailabilityFilter,
    ]

    readonly_fields = ["api_info"]

    actions = [ping_topics, garbage_collect]

    @admin.display(description="Consumer(s)")  # type: ignore
    def get_consumers(self, obj: models.BrokerTopic) -> str:
        return ", ".join(map(str, obj.consumers))[:60]

    @admin.display(description="Producer(s)")  # type: ignore
    def get_producers(self, obj: models.BrokerTopic) -> str:
        return ", ".join(map(str, obj.producers))[:60]

    @admin.display(boolean=True)  # type: ignore
    def is_response_topic(self, obj: models.BrokerTopic) -> bool:
        return obj.is_response_topic

    @admin.display(description="Online?", boolean=True)  # type: ignore
    def get_availability(self, obj: models.BrokerTopic) -> Optional[bool]:
        return obj.availability

    def pending_messages(self, obj: models.BrokerTopic) -> str:
        """Get the number of messages that have not yet been delivered"""
        return str(obj.brokermessage_set.filter(delivered_to=None).count())

    def stored_messages(self, obj: models.BrokerTopic) -> str:
        """Get the number of messages in the database."""
        return str(obj.brokermessage_set.count())

    def stored_responses(self, obj: models.BrokerTopic) -> str:
        """Get the number of messages in the database."""
        if obj.is_response_topic:
            return "-"
        else:
            return str(
                models.BrokerMessage.objects.filter(
                    topic__responsetopic__is_response_for=obj
                ).count()
            )


@admin.register(models.ResponseTopic)
class ResponseTopicAdmin(BrokerTopicAdmin):
    """An admin for response topics."""

    search_fields = BrokerTopicAdmin.search_fields + [
        "is_response_for__slug__icontains"
    ]

    list_display = BrokerTopicAdmin.list_display[:-1]

    list_filter = [
        "is_public",
        "date_created",
    ]


@admin.action(description="Send messages")
def send_messages(modeladmin, request, queryset):
    for message in queryset:
        message.send()


@admin.action(description="Ping topics")
def ping_message_topics(modeladmin, request, queryset):
    pks = queryset.values_list("topic__pk", flat=True)
    for topic in models.BrokerTopic.objects.filter(
        pk__in=pks, supports_dasf=True
    ):
        topic.ping()


class IsResponseMessageFilter(admin.SimpleListFilter):
    """Filter broker messages to a response topic"""

    title = "is response message"

    parameter_name = "topic"

    def lookups(self, request: Any, model_admin: Any) -> List[Tuple[Any, str]]:
        return [("1", "Yes"), ("0", "No")]

    def queryset(
        self, request: Any, queryset: QuerySet[Any]
    ) -> Optional[QuerySet[Any]]:
        value = self.value()
        if value == "1":
            return queryset.filter(topic__responsetopic__isnull=False)
        elif value == "0":
            return queryset.filter(topic__responsetopic__isnull=True)
        else:
            return queryset


class HasBeenDeliveredFilter(admin.SimpleListFilter):
    """Filter broker messages by their delivery state"""

    title = "has been delivered"

    parameter_name = "delivered_to"

    def lookups(self, request: Any, model_admin: Any) -> List[Tuple[Any, str]]:
        return [("1", "Yes"), ("0", "No")]

    def queryset(
        self, request: Any, queryset: QuerySet[Any]
    ) -> Optional[QuerySet[Any]]:
        value = self.value()
        if value == "1":
            return queryset.filter(delivered_to=True)
        elif value == "0":
            return queryset.filter(delivered_to=None)
        else:
            return queryset


@admin.register(models.BrokerMessage)
class BrokerMessageAdmin(GuardedModelAdmin):
    """An admin for a broker message"""

    search_fields = [
        "topic__slug__icontains",
        "topic__responsetopic__is_response_for__slug__icontains",
        "user__username__icontains",
        "user__email__icontains",
        "user__first_name__icontains",
        "user__last_name__icontains",
    ]

    list_display = [
        "message_id",
        "topic",
        "topic_availability",
        "is_response",
        "user",
        "delivered",
        "date_created",
    ]

    list_filter = [
        "date_created",
        IsResponseMessageFilter,
        HasBeenDeliveredFilter,
    ]

    readonly_fields = ["content"]

    actions = [send_messages, ping_message_topics]

    @admin.display(boolean=True)  # type: ignore
    def delivered(self, obj: models.BrokerMessage) -> bool:
        return obj.delivered_to.exists()

    @admin.display(boolean=True)  # type: ignore
    def is_response(self, obj: models.BrokerMessage) -> bool:
        return obj.is_response

    @admin.display(boolean=True, description="Topic online?")
    def topic_availability(self, obj: models.BrokerMessage) -> Optional[bool]:
        return obj.topic.availability



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Models
------

Consumers for the dasf-broker-django app.
"""

from __future__ import annotations

import datetime as dt
import json
from typing import TYPE_CHECKING, Dict, Optional, cast
from uuid import uuid4

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer
from django.conf import settings
from django.urls import set_script_prefix
from django.utils.timezone import now

from dasf_broker import app_settings, constants

if TYPE_CHECKING:
    from dasf_broker.models import BrokerTopic


class TopicProducer(JsonWebsocketConsumer):
    """A producer of messages for a certain topic."""

    @property
    def dasf_topic(self) -> Optional[BrokerTopic]:
        from dasf_broker.models import BrokerTopic

        return BrokerTopic.objects.filter(slug=self.dasf_topic_slug).first()

    def connect(self):
        from guardian.shortcuts import get_anonymous_user

        try:
            user = self.scope["user"]
        except KeyError:
            self.user = get_anonymous_user()
        else:
            self.user = get_anonymous_user() if user.is_anonymous else user
        self.dasf_topic_slug = self.scope["url_route"]["kwargs"]["slug"]

        self.accept()

    def receive_json(self, content: Dict):
        """Distibute the message to the consumers."""
        from guardian.shortcuts import assign_perm

        from dasf_broker.models import BrokerTopic

        # HACK: daphne seems to not take the FORCE_SCRIPT_NAME into account.
        if getattr(settings, "FORCE_SCRIPT_NAME", None):
            set_script_prefix(settings.FORCE_SCRIPT_NAME)  # type: ignore

        message_id = content.setdefault("messageId", str(uuid4()))

        ack = {
            "messageId": message_id,
        }

        if "context" in content:
            ack["context"] = content["context"]

        topic = self.dasf_topic

        if topic is None:
            if (
                app_settings.DASF_CREATE_TOPIC_ON_MESSAGE
                and self.user.has_perm("dasf_broker.add_brokertopic")
            ):
                topic = BrokerTopic.objects.create(slug=self.dasf_topic_slug)
                assign_perm("can_produce", self.user, topic)
                assign_perm("can_consume", self.user, topic)
            elif not app_settings.DASF_CREATE_TOPIC_ON_MESSAGE:
                ack["result"] = f"Topic {self.dasf_topic_slug} does not exist."
            else:
                ack["result"] = "User is not allowed to create topics."

        if topic is not None and (
            topic.is_public
            or self.user.has_perm("dasf_broker.can_produce", topic)
        ):
            # create a topic for the response
            topic = cast(BrokerTopic, topic)

            topic.create_and_send_message(self.user, content)

            self.post_response_message(content)

            # acknowledge the message
            ack["result"] = "ok"
        elif topic is not None:
            ack["result"] = "User is not allowed to publish to this topic"

        self.send_json(ack)

    def post_response_message(self, content):
        """Hook to handle a message.

        This method is supposed to be implemented by subclasses for a response
        message."""
        pass


class PongConsumer(JsonWebsocketConsumer):
    """A consumer to handle pong messages of a topic."""

    def connect(self):
        from guardian.shortcuts import get_anonymous_user

        try:
            user = self.scope["user"]
        except KeyError:
            self.user = get_anonymous_user()
        else:
            self.user = get_anonymous_user() if user.is_anonymous else user

        self.accept()

    def receive_json(self, content, **kwargs):
        from dasf_broker.models import BrokerTopic

        # HACK: daphne seems to not take the FORCE_SCRIPT_NAME into account.
        if getattr(settings, "FORCE_SCRIPT_NAME", None):
            set_script_prefix(settings.FORCE_SCRIPT_NAME)  # type: ignore

        message_id = content.setdefault("messageId", str(uuid4()))

        ack = {
            "messageId": message_id,
        }
        if "context" in content:
            ack["context"] = content["context"]

        topic_slug = content.get("properties", {}).get("source_topic")

        if not topic_slug:
            ack["result"] = "No source topic found."
            self.send_json(ack)
            return
        else:
            topic = BrokerTopic.objects.filter(slug=topic_slug).first()

            if topic is None or not self.user.has_perm("can_consume", topic):
                ack["result"] = "User is not allowed to publish to this topic"
                self.send_json(ack)
                return

        props = content.get("properties", {})
        message_type = props.get("messageType")
        if message_type == constants.MessageType.PONG:
            topic.last_pong = now()
            topic.save()
        elif (
            message_type == constants.MessageType.RESPONSE
            and "api_info" in props
        ):
            topic.api_info = json.loads(props["api_info"])
            topic.save()

        ack["result"] = "ok"

        self.send_json(ack)


class TopicConsumer(JsonWebsocketConsumer):
    """A consumer of messages."""

    @property
    def dasf_topic(self) -> Optional[BrokerTopic]:
        from dasf_broker.models import BrokerTopic

        return BrokerTopic.objects.filter(slug=self.dasf_topic_slug).first()

    def connect(self):
        from guardian.shortcuts import get_anonymous_user

        try:
            user = self.scope["user"]
        except KeyError:
            self.user = get_anonymous_user()
        else:
            self.user = get_anonymous_user() if user.is_anonymous else user
        self.dasf_topic_slug = slug = self.scope["url_route"]["kwargs"]["slug"]
        self.room_group_name = f"dasf_topic_{slug}"

        # Join room group
        async_to_sync(self.channel_layer.group_add)(
            self.room_group_name, self.channel_name
        )

        self.accept()

        # send the messages that have not yet been delivered
        topic = self.dasf_topic
        if topic is not None and self.user.has_perm("can_consume", topic):
            messages = topic.get_outstanding_messages(self.user)
            topic.garbage_collect_on = None
            if messages:
                topic.last_ping = now()
                topic.save()
                for message in messages:
                    self.send_json(message.content)
            else:
                self.dasf_api_info({"type": "dasf.api_info"})

    def receive_json(self, content, **kwargs):
        from dasf_broker.models import BrokerMessage

        # HACK: daphne seems to not take the FORCE_SCRIPT_NAME into account.
        if getattr(settings, "FORCE_SCRIPT_NAME", None):
            set_script_prefix(settings.FORCE_SCRIPT_NAME)  # type: ignore

        topic = self.dasf_topic

        if topic is None or not self.user.has_perm("can_consume", topic):
            return

        topic.last_pong = now()
        topic.save()

        choices = topic.StoreMessageChoices

        if topic.effective_store_messages in [choices.STORE, choices.CACHEALL]:
            # mark the message as delivered
            try:
                message = topic.brokermessage_set.get(
                    message_id=content["messageId"]
                )
            except BrokerMessage.DoesNotExist:
                pass
            else:
                message.delivered_to.add(self.user)
                if (
                    topic.effective_store_messages == choices.CACHEALL
                    and message.delivered_to_all
                ):
                    message.delete()
        elif topic.effective_store_messages == choices.CACHE:
            topic.brokermessage_set.filter(
                message_id=content["messageId"]
            ).delete()

    def disconnect(self, close_code):
        # Leave room group
        async_to_sync(self.channel_layer.group_discard)(
            self.room_group_name, self.channel_name
        )
        topic = self.dasf_topic
        if (
            topic
            and topic.is_response_topic
            and (
                topic.effective_store_messages == "disabled"
                or (
                    topic.effective_store_messages == "cache"
                    and not topic.get_outstanding_messages()
                )
            )
        ):
            # delete the response topic as it is not needed anymore
            topic.garbage_collect_on = now() + dt.timedelta(minutes=10)
            topic.save()
        else:
            topic.ping()

    def dasf_message(self, event):
        topic = self.dasf_topic
        if topic is not None and self.user.has_perm("can_consume", topic):
            topic.last_ping = now()
            topic.save()
            self.send_json(event["content"])

    def dasf_ping(self, event):
        """Send a ping message to the topic."""
        topic = self.dasf_topic
        if not topic.supports_dasf:
            # do nothing
            return
        date_pinged = now()
        content = {
            "properties": {
                "requestContext": 1,
                "messageType": constants.MessageType.PING,
                "response_topic": "status-ping",
            },
            "publishTime": date_pinged.isoformat(),
            "messageId": str(uuid4()),
            "payload": "",
        }
        topic.last_ping = date_pinged
        topic.save()
        event["content"] = content
        event["type"] = "dasf_message"
        self.dasf_message(event)

    def dasf_api_info(self, event):
        """Send a ping message to the topic."""
        topic = self.dasf_topic
        if not topic.supports_dasf:
            # do nothing
            return
        content = {
            "properties": {
                "requestContext": 1,
                "messageType": constants.MessageType.API_INFO,
                "response_topic": "status-ping",
            },
            "publishTime": now().isoformat(),
            "messageId": str(uuid4()),
            "payload": "",
        }
        event["content"] = content
        event["type"] = "dasf_message"
        self.dasf_message(event)



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Ping DASF backend modules."""

from django_extensions.management.jobs import MinutelyJob


class Job(MinutelyJob):
    help = "Ping DASF Backend Modules."

    def execute(self):
        from dasf_broker.models import BrokerTopic

        for topic in BrokerTopic.objects.filter(
            responsetopic__isnull=True, supports_dasf=True
        ):
            topic.ping()



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Constants within the DASF Framework
-----------------------------------

Constants from :mod:`demessaging.PulsarMessageConstants`.
"""

from enum import Enum


class PropertyKeys(str, Enum):
    REQUEST_CONTEXT = "requestContext"
    RESPONSE_TOPIC = "response_topic"
    REQUEST_MESSAGEID = "requestMessageId"
    MESSAGE_TYPE = "messageType"
    FRAGMENT = "fragment"
    NUM_FRAGMENTS = "num_fragments"
    STATUS = "status"


class MessageType(str, Enum):
    """Supported message types."""

    PING = "ping"
    PONG = "pong"
    REQUEST = "request"
    RESPONSE = "response"
    LOG = "log"
    INFO = "info"
    PROGRESS = "progress"
    API_INFO = "api_info"



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""URL config
----------

URL patterns of the dasf-broker-django to be included via::

    from django.urls import include, path

    urlpatters = [
        path(
            "dasf-broker-django",
            include("dasf_broker.urls"),
        ),
    ]
"""

from __future__ import annotations

from typing import Any

from django.urls import path  # noqa: F401

from dasf_broker import views  # noqa: F401

#: App name for the dasf-broker-django to be used in calls to
#: :func:`django.urls.reverse`
app_name = "dasf_broker"

#: urlpattern for the Helmholtz AAI
urlpatterns: list[Any] = [
    path(
        "<slug>/status/",
        views.BrokerTopicStatusView.as_view(),
        name="brokertopic-status",
    ),
    path(
        "<slug>/ping/",
        views.BrokerTopicPingView.as_view(),
        name="brokertopic-ping",
    ),
]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

from typing import Any, List

from channels.routing import URLRouter
from django.urls import path, re_path

from dasf_broker import app_settings

from . import consumers

websocket_urlpatterns: List[Any] = [
    path(
        app_settings.DASF_WEBSOCKET_URL_ROUTE,
        URLRouter(
            [
                re_path(r"^status-ping/?$", consumers.PongConsumer.as_asgi()),
                re_path(
                    r"^(?P<slug>[-_\w]+)/?$",
                    consumers.TopicProducer.as_asgi(),
                ),
                re_path(
                    r"^(?P<slug>[-_\w]+)/(?P<subscription>[-\.:\w]+)/?$",
                    consumers.TopicConsumer.as_asgi(),
                ),
            ]
        ),
    )
]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Management command to edit or create a Broker topic"""

import argparse
from typing import Literal, Optional

from django.core.management.base import BaseCommand
from guardian.shortcuts import assign_perm, get_anonymous_user, remove_perm


class Command(BaseCommand):
    """Django command to migrate the database."""

    help = "Create or update a broker topic."

    def add_arguments(self, parser):
        """Add connection arguments to the parser."""

        parser.add_argument("topic_slug", help="The slug for the topic")

        parser.add_argument(
            "-n", "--new", help="Create a new topic.", action="store_true"
        )

        if hasattr(argparse, "BooleanOptionalAction"):
            parser.add_argument(
                "--public",
                action=argparse.BooleanOptionalAction,
                help="Make the topic public.",
                default=None,
            )

            parser.add_argument(
                "--dasf",
                action=argparse.BooleanOptionalAction,
                help=(
                    "Does the consumer of the topic support the DASF "
                    "messaging protocoll?"
                ),
                dest="supports_dasf",
                default=None,
            )

        else:
            # python 3.8
            parser.add_argument(
                "--public",
                action="store_true",
                help="Make the topic public.",
                default=None,
            )
            parser.add_argument(
                "--no-public",
                action="store_false",
                dest="public",
                help="Make the topic not public.",
                default=None,
            )

            parser.add_argument(
                "--dasf",
                action="store_true",
                help=(
                    "Does the consumer of the topic support the DASF "
                    "messaging protocoll?"
                ),
                dest="supports_dasf",
                default=None,
            )

            parser.add_argument(
                "--no-dasf",
                action="store_false",
                help=(
                    "Use this option if the consumer does not support the "
                    "DASF protocoll."
                ),
                dest="supports_dasf",
                default=None,
            )

        parser.add_argument(
            "--anonymous",
            action="store_true",
            help=(
                "Make the topic available for anonymous consumers and "
                "producers. This option involves '--public' and "
                "'-c AnonymousUser'."
            ),
        )

        parser.add_argument(
            "-c",
            "--consumer",
            help=(
                "Username of the user account that you want to make a "
                "consumer of the topic"
            ),
        )

        parser.add_argument(
            "-p",
            "--producer",
            help=(
                "Username of the user account that you want to make a "
                "producer of the topic"
            ),
        )

        parser.add_argument(
            "--store-messages",
            help=(
                "Should messages be store for this topic? When using "
                "'default', the value of the DASF_STORE_MESSAGES setting is "
                "used."
            ),
            choices=[
                "default",
                "disabled",
                "cache",
                "cacheall",
                "store",
            ],
        )

        parser.add_argument(
            "-vs",
            "--view-status",
            help=(
                "Username of the user account that should be able to view "
                "the status of the topic"
            ),
        )

        parser.add_argument(
            "--create-consumer",
            help=(
                "Create a new user account for the consumer with the given "
                "username"
            ),
            action="store_true",
        )

        parser.add_argument(
            "--create-producer",
            help=(
                "Create a new user account for the producer with the given "
                "username"
            ),
            action="store_true",
        )

        parser.add_argument(
            "--create-view-status",
            help=(
                "Create a new user account for the user that should be able "
                " to view the status with the given username"
            ),
            action="store_true",
        )

        parser.add_argument(
            "-rm",
            "--remove-permissions",
            help=(
                "Remove the permissions for the given consumer and/or "
                "producer instead of assigning."
            ),
            action="store_true",
        )

        parser.add_argument(
            "-i",
            "--show-info",
            action="store_true",
            help=(
                "Show information on the topic (including a list of the user "
                "names of consumers and producers and whether the topic is "
                "public or not. The updates to the topic are done before "
                "printing out the info."
            ),
        )

        parser.add_argument(
            "-db",
            "--database",
            help=(
                "The Django database identifier (see settings.py), "
                "default: %(default)s"
            ),
            default="default",
        )

    def handle(  # type: ignore
        self,
        topic_slug: str,
        *args,
        new: bool = False,
        public: Optional[bool] = None,
        supports_dasf: Optional[bool] = None,
        store_messages: Optional[
            Literal["default", "disabled", "cache", "cacheall", "store"]
        ] = None,
        anonymous: bool = False,
        consumer: Optional[str] = None,
        producer: Optional[str] = None,
        view_status: Optional[str] = None,
        create_consumer: bool = False,
        create_producer: bool = False,
        create_view_status: bool = False,
        remove_permissions: bool = False,
        show_info: bool = False,
        database: str = "default",
        **options,
    ):
        """Migrate the database."""

        from django.contrib.auth import get_user_model
        from django.db import IntegrityError

        from dasf_broker.models import BrokerTopic

        User = get_user_model()  # type: ignore  # noqa: F811

        if new:
            try:
                topic = BrokerTopic.objects.using(database).create(
                    slug=topic_slug,
                )
            except IntegrityError:
                raise ValueError(
                    f"A topic with the slug {topic_slug} already exists."
                )
        else:
            try:
                topic = BrokerTopic.objects.using(database).get(
                    slug=topic_slug,
                )
            except BrokerTopic.DoesNotExist:
                raise ValueError(
                    f"A topic with the slug {topic_slug} does not exist. "
                    "If you want to create it, please use "
                    "'--new' option instead of '--update'."
                )

        if anonymous:
            topic.is_public = True
            assign_perm("can_consume", get_anonymous_user(), topic)
            topic.save()

        if public is not None:
            topic.is_public = public
            topic.save()

        if supports_dasf is not None:
            topic.supports_dasf = supports_dasf
            topic.save()

        if store_messages is not None:
            if store_messages == "default":
                topic.store_messages = None
            else:
                topic.store_messages = store_messages
            topic.save()

        if consumer is not None:
            if create_consumer:
                try:
                    user = User.objects.create(username=consumer)
                except IntegrityError:
                    raise ValueError(
                        f"A user with the username {consumer} already exists!"
                    )
            else:
                try:
                    user = User.objects.get(username=consumer)
                except User.DoesNotExist:
                    raise ValueError(
                        f"A user with the username {consumer} does not exist. "
                        "If you want to create it, please add the "
                        "'--create-consumer' option."
                    )
            if remove_permissions:
                remove_perm("can_consume", user, topic)
            else:
                assign_perm("can_consume", user, topic)

        if producer is not None:
            create_producer = create_producer and not (
                create_consumer and consumer == producer
            )
            if create_producer:
                try:
                    user = User.objects.create(username=producer)
                except IntegrityError:
                    raise ValueError(
                        f"A user with the username {producer} already exists!"
                    )
            else:
                try:
                    user = User.objects.get(username=producer)
                except User.DoesNotExist:
                    raise ValueError(
                        f"A user with the username {producer} does not exist. "
                        "If you want to create it, please add the "
                        "'--create-producer' option."
                    )
            if remove_permissions:
                remove_perm("can_produce", user, topic)
            else:
                assign_perm("can_produce", user, topic)

        if view_status is not None:
            create_view_status = create_view_status and not (
                (create_consumer and consumer == view_status)
                or (create_producer and producer == view_status)
            )
            if create_view_status:
                try:
                    user = User.objects.create(username=view_status)
                except IntegrityError:
                    raise ValueError(
                        f"A user with the username {view_status} already exists!"
                    )
            else:
                try:
                    user = User.objects.get(username=view_status)
                except User.DoesNotExist:
                    raise ValueError(
                        f"A user with the username {view_status} does not exist. "
                        "If you want to create it, please add the "
                        "'--create-view-status' option."
                    )
            if remove_permissions:
                remove_perm("can_view_status", user, topic)
            else:
                assign_perm("can_view_status", user, topic)

        if show_info:
            print("Consumers:")
            for user in topic.consumers:
                print(f"    {user}")
            print("public:")
            print(f"    {topic.is_public}")
            print("Producers:")
            for user in topic.producers:
                print(f"    {user}")
            print("Users that can view the status:")
            for user in topic.status_viewers:
                print(f"    {user}")



# SPDX-FileCopyrightText: 2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

# Generated by Django 3.2.25 on 2025-04-28 19:24

import django.core.serializers.json
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("dasf_broker", "0002_auto_20230213_1501"),
    ]

    operations = [
        migrations.AddField(
            model_name="brokertopic",
            name="api_info",
            field=models.JSONField(
                default=dict,
                encoder=django.core.serializers.json.DjangoJSONEncoder,
            ),
        ),
    ]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Django-based Message Broker for DASF

A Django-based message broker for the Data Analytics Software Framework (DASF)
"""

from __future__ import annotations

from . import _version

__version__ = _version.get_versions()["version"]

__author__ = "Philipp S. Sommer"
__copyright__ = "Copyright (C) 2022 Helmholtz-Zentrum Hereon"
__credits__ = [
    "Philipp S. Sommer",
]
__license__ = "EUPL-1.2"

__maintainer__ = "Helmholtz-Zentrum Hereon"
__email__ = "hcdc_support@hereon.de"

__status__ = "Pre-Alpha"



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

# Generated by Django 3.2.15 on 2023-02-13 15:01

from django.conf import settings
from django.db import migrations, models


def migrate_delivered(apps, schema_editor):
    """Migrate the delivered field to delivered_to."""
    BrokerMessage = apps.get_model("dasf_broker", "BrokerMessage")
    User = apps.get_model(*settings.AUTH_USER_MODEL.split("."))

    queryset = BrokerMessage.objects.filter(delivered=True)
    if queryset:
        username = getattr(settings, "ANONYMOUS_USER_NAME", "AnonymousUser")
        user = User.objects.get(username=username)
        for message in queryset:
            message.delivered_to.add(user)


def undo_migrate_delivered(apps, schema_editor):
    """Undo the migration of the delivered field to delivered_to."""

    BrokerMessage = apps.get_model("dasf_broker", "BrokerMessage")

    queryset = BrokerMessage.objects.filter(delivered_to__pk__isnull=False)
    queryset.update(delivered=True)


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("dasf_broker", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="brokermessage",
            name="delivered_to",
            field=models.ManyToManyField(
                help_text="To whom has this message been delivered already?",
                related_name="brokermessage_set_delivered",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.RunPython(migrate_delivered, undo_migrate_delivered),
        migrations.RemoveField(
            model_name="brokermessage",
            name="delivered",
        ),
        migrations.AddField(
            model_name="brokertopic",
            name="supports_dasf",
            field=models.BooleanField(
                default=True,
                help_text="Does this topic support the DASF messaging protocoll?",
            ),
        ),
        migrations.AddField(
            model_name="brokertopic",
            name="garbage_collect_on",
            field=models.DateTimeField(
                blank=True,
                help_text="Specify a time when this topic should be removed from the database.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="brokertopic",
            name="store_messages",
            field=models.CharField(
                blank=True,
                choices=[
                    (
                        None,
                        "Use the server default",
                    ),
                    ("disabled", "Do not store messages at all"),
                    (
                        "cache",
                        "Cache until it has been delivered to at least on consumer",
                    ),
                    (
                        "cacheall",
                        "Cache until it has been delivered to all consumers",
                    ),
                    ("store", "Store messages forever"),
                ],
                help_text="Should messages be stored?",
                max_length=20,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="brokertopic",
            name="last_ping",
            field=models.DateTimeField(
                blank=True,
                help_text="When has the topic last been pinged?",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="brokertopic",
            name="last_pong",
            field=models.DateTimeField(
                blank=True,
                help_text="When has the topic last replyed on a ping?",
                null=True,
            ),
        ),
        migrations.AlterModelOptions(
            name="brokertopic",
            options={
                "permissions": (
                    (
                        "can_produce",
                        "Can publish messages to the topic (Producer).",
                    ),
                    (
                        "can_consume",
                        "Can consume messages to the topic (Consumer).",
                    ),
                    (
                        "can_view_status",
                        "Can view the status of the consumer.",
                    ),
                )
            },
        ),
    ]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""Template tags
-------------

This module contains the :mod:`~dasf_broker.templatetags.dasf` module that
defines template tags that should be used within DASF.
"""



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""App config
----------

App config for the dasf_broker app.
"""

from django.apps import AppConfig


class DjangoHelmholtzAaiConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "dasf_broker"



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""URL config
----------

URL patterns of the dasf-broker-django frontend to be included via::

    from django.urls import include, path

    urlpatters = [
        path(
            "dasf",
            include("dasf_broker.frontend.urls"),
        ),
    ]
"""

from __future__ import annotations

from typing import Any

from django.urls import path  # noqa: F401

from . import views  # noqa: F401

app_name = "dasf_frontend"

urlpatterns: list[Any] = [
    path(
        "<slug>/",
        views.BrokerTopicFrontendView.as_view(),
        name="brokertopic-frontend",
    ),
]



# SPDX-FileCopyrightText: 2022-2025 Helmholtz-Zentrum hereon GmbH
#
# SPDX-License-Identifier: EUPL-1.2

"""App settings
------------

This module defines the settings options for the
``dasf-broker-django`` app.
"""

from __future__ import annotations

from enum import Enum
from typing import Optional

from django.conf import settings  # noqa: F401

#: Create new topics on message
#:
#: This flag controls if new topics are created when a message comes from a
#: producer. If False, messages with non-existing topics are ignored.
#:
#: Note that a user also needs the `dasf_broker.add_BrokerTopic`
#: permission to create topics.
DASF_CREATE_TOPIC_ON_MESSAGE: bool = getattr(
    settings, "DASF_CREATE_TOPIC_ON_MESSAGE", True
)


class StoreMessageOptions(str, Enum):
    DISABLED = "disabled"
    CACHE = "cache"
    CACHEALL = "cacheall"
    STORE = "store"


#: Shall the messages be stored?
#:
#: This flag controls whether the message broker caches messages from producers
#: until they are consumed by the consumer. This is useful if the consumer
#: looses connection to the server. This settings can take three different
#: values:
#:
#: ``"disabled"``
#:     The message is not stored at all
#: ``"cache"``
#:     The message is stored and removed ones one of the potential `consumers`
#:     acknowledges the message
#: ``"cacheall"``
#:     The message is stored and removed ones **all** `consumers` acknowledged
#:     the message
#: ``"store"``
#:     The message and response topics are stored forever and are not
#:     automatically removed
#:
#: See Also
#: --------
#: DASF_STORE_SOURCE_MESSAGES
#: DASF_STORE_RESPONSE_MESSAGES
DASF_STORE_MESSAGES: StoreMessageOptions = StoreMessageOptions(
    getattr(settings, "DASF_STORE_MESSAGES", StoreMessageOptions.CACHE)
)

#: Shall the source messages be stored?
#:
#: This flag controls whether the message broker caches messages from producers
#: to topics that are **not** marked as response topic. If this setting is not
#: set, we use the ``DASF_STORE_MESSAGES`` setting.
DASF_STORE_SOURCE_MESSAGES: StoreMessageOptions = StoreMessageOptions(
    getattr(settings, "DASF_STORE_SOURCE_MESSAGES", DASF_STORE_MESSAGES)
)

#: Shall the messages to response topics be stored?
#:
#: This flag controls whether the message broker caches messages from producers
#: to topics that are marked as response topic. If this setting is not set, we
#: use the ``DASF_STORE_MESSAGES`` setting.
DASF_STORE_RESPONSE_MESSAGES: StoreMessageOptions = StoreMessageOptions(
    getattr(settings, "DASF_STORE_RESPONSE_MESSAGES", DASF_STORE_MESSAGES)
)

#: URL route for the websocket
#:
#: This setting controls, where we expect to find the websockets.
#: As there is no analog to :func:`django.urls.reverse` for channels, you
#: should use this setting in your ``asgi.py`` file to include the routes
#: of this package.
#:
#: Examples
#: --------
#: In your ``asgi.py`` file, include it like:
#:
#: .. code-block:: python
#:
#:     from channels.routing import ProtocolTypeRouter, URLRouter
#:     from django.core.asgi import get_asgi_application
#:     from channels.auth import AuthMiddlewareStack
#:     import dasf_broker.routing as dasf_routing
#:     from dasf_broker.app_settings import DASF_WEBSOCKET_URL_ROUTE
#:
#:     application = ProtocolTypeRouter(
#:         {
#:             "http": get_asgi_application(),
#:             "websocket": AuthMiddlewareStack(
#:                 URLRouter(
#:                     [
#:                         path(
#:                             DASF_WEBSOCKET_URL_ROUTE,
#:                             URLRouter(dasf_routing.websocket_urlpatterns),
#:                         )
#:                     ]
#:                 )
#:             ),
#:         }
#:     )
DASF_WEBSOCKET_URL_ROUTE: str = getattr(
    settings, "DASF_WEBSOCKET_URL_ROUTE", "ws/"
)


#: root URL to your application
#:
#: You can use this setting if you are behind a reverse proxy and the
#: host names, etc. are not handled correctly.
#:
#: If you leave this empty, we will use the ``build_absolute_uri`` method
#: of the http request.
#:
#: Examples
#: --------
#: A standard value for this would be ``http://localhost:8000``
ROOT_URL: Optional[str] = None
